# coding: utf-8
"""
    flask.ext.makestatic.watcher
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2013 by Daniel Neuhäuser
    :license: BSD, see LICENSE.rst for details
"""
import os
import time
import errno
import threading

from flask.ext.makestatic._compat import iteritems


class Signal(object):
    def __init__(self):
        self.listeners = []

    def send(self, *args, **kwargs):
        for listener in self.listeners:
            listener(*args, **kwargs)

    def connect(self, function):
        self.listeners.append(function)
        return function


class Watcher(object):
    def __init__(self):
        self.files = {}
        self.directories = {}
        self._lock = threading.RLock()

        self.file_added = Signal()
        self.file_modified = Signal()
        self.file_removed = Signal()

        self.directory_added = Signal()
        self.directory_modified = Signal()
        self.directory_removed = Signal()

        self._stopped = False

    def _listdir(self, directory):
        return set(os.path.join(directory, f) for f in os.listdir(directory))

    def add_file(self, file):
        with self._lock:
            self.files[file] = os.stat(file).st_mtime

    def add_directory(self, directory, ignore_contained=True):
        with self._lock:
            paths = self._listdir(directory)
            if ignore_contained:
                self.directories[directory] = paths
            else:
                self.directories[directory] = set()
            for path in paths:
                if os.path.isfile(path):
                    self.add_file(path)
                elif os.path.isdir(path):
                    self.add_directory(path, ignore_contained=ignore_contained)

    def stop(self):
        self._stopped = True

    def watch(self, sleep=0.1):
        while not self._stopped:
            with self._lock:
                new_directories = []
                removed_directories = []
                for directory, seen in iteritems(self.directories):
                    try:
                        current = self._listdir(directory)
                    except OSError as error:
                        if error.errno == errno.ENOENT:
                            self.directory_removed.send(directory)
                            removed_directories.append(directory)
                        else:
                            raise
                    else:
                        for path in seen ^ current:
                            self.directory_modified.send(directory)
                            if path in current:
                                if os.path.isfile(path):
                                    self.add_file(path)
                                    self.file_added.send(path)
                                elif os.path.isdir(path):
                                    new_directories.append(path)
                                    self.directory_added.send(path)
                            else:
                                if os.path.isfile(path):
                                    self.file_removed.send(path)
                                    self.files.pop(path)
                                elif os.path.isdir(path):
                                    self.directory_removed.send(path)
                        self.directories[directory] = current
                for directory in new_directories:
                    self.add_directory(directory, ignore_contained=False)
                for directory in removed_directories:
                    del self.directories[directory]
                removed_files = []
                for file, last_known_modified_time in iteritems(self.files):
                    try:
                        modified_time = os.stat(file).st_mtime
                    except OSError as error:
                        if error.errno == errno.ENOENT:
                            self.file_removed.send(file)
                            removed_files.append(file)
                        else:
                            raise
                    else:
                        if modified_time > last_known_modified_time:
                            self.files[file] = modified_time
                            self.file_modified.send(file)
                            directory = os.path.dirname(file)
                            if directory in self.directories:
                                self.directory_modified.send(directory)
                for file in removed_files:
                    del self.files[file]
                time.sleep(sleep)


class ThreadingMixin(object):
    def watch(self, sleep=0.1):
        thread = threading.Thread(target=super(ThreadingMixin, self).watch,
                                  kwargs=dict(sleep=sleep))
        thread.daemon = True
        thread.start()


class ThreadedWatcher(ThreadingMixin, Watcher):
    pass



# coding: utf-8
"""
    flask.ext.makestatic._compat
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2013 by Daniel Neuhäuser
    :license: BSD, see LICENSE.rst for details
"""
import sys


PY2 = sys.version_info[0] == 2


if PY2:
    def iteritems(d):
        return d.iteritems()

    from cStringIO import StringIO
else:
    def iteritems(d):
        return d.items()

    from io import StringIO


__all__ = ['PY2', 'iteritems', 'StringIO']



# coding: utf-8
import os
import sys
import time
import atexit
import shutil
import tempfile
import unittest
import threading
from warnings import catch_warnings
from contextlib import closing, contextmanager

from flask import Flask
from werkzeug.exceptions import NotFound

from flask.ext.makestatic import MakeStatic, RuleMissing
from flask.ext.makestatic._compat import StringIO
from flask.ext.makestatic.watcher import ThreadedWatcher


TEST_APPS = os.path.join(os.path.dirname(__file__), 'test_apps')
sys.path.insert(0, TEST_APPS)


def get_temporary_filename(directory=None):
    filename = tempfile.mkstemp(dir=directory)[1]
    @atexit.register
    def remove_file():
        try:
            os.remove(filename)
        except EnvironmentError:
            pass
    return filename


def get_temporary_directory():
    directory = tempfile.mkdtemp()
    @atexit.register
    def remove_directory():
        try:
            shutil.rmtree(directory)
        except EnvironmentError:
            pass
    return directory


def bump_modification_time(path):
    stat = os.stat(path)
    os.utime(path, (stat.st_atime, stat.st_mtime + 1))


@contextmanager
def catch_stdout():
    old_stdout = sys.stdout
    sys.stdout = rv = StringIO()
    try:
        yield rv
    finally:
        sys.stdout = old_stdout


class MakeStaticTestCase(unittest.TestCase):
    def tearDown(self):
        for test_app_dir in os.listdir(TEST_APPS):
            static_dir = os.path.join(TEST_APPS, test_app_dir, 'static')
            if os.path.isdir(static_dir):
                for root, dirs, files in os.walk(static_dir):
                    for file in files:
                        if file == '.gitignore':
                            continue
                        os.remove(os.path.join(root, file))
                    for dir in dirs:
                        os.rmdir(os.path.join(root, dir))

    @contextmanager
    def make_static(self, import_name):
        app = Flask(import_name)
        make_static = MakeStatic(app)
        watcher = make_static.watch(sleep=0.01)
        try:
            yield app, make_static
        finally:
            watcher.stop()

    def test_send_static_file(self):
        with self.make_static('working') as (app, make_static):
            with app.test_request_context():
                with closing(app.send_static_file('foo')) as response:
                    self.assertEqual(response.status_code, 200)

                with closing(app.send_static_file('spam')) as response:
                    self.assertEqual(response.status_code, 200)

                with closing(app.send_static_file('bar')) as response:
                    self.assertEqual(response.status_code, 200)
                    response.direct_passthrough = False
                    self.assertEqual(response.data, b'abc\ndef\n')

                self.assertRaises(NotFound, app.send_static_file, 'baz')

                with closing(app.send_static_file('eggs.css')) as response:
                    self.assertEqual(response.status_code, 200)

                with catch_warnings(record=True): # silence RuleMissing warning
                    with catch_stdout() as stdout:
                        filename = os.path.basename(get_temporary_filename(
                            directory=make_static.assets_folder
                        ))
                        time.sleep(0.05)

                self.assertEqual(
                    stdout.getvalue(),
                    'Flask-MakeStatic: detected new asset %s, compiling\n' %
                    filename
                )

                self.assertRaises(NotFound, app.send_static_file, filename)

                with catch_stdout() as stdout:
                    bump_modification_time(
                        os.path.join(make_static.assets_folder, 'foo')
                    )
                    time.sleep(0.05)

                self.assertEqual(
                    stdout.getvalue(),
                    'Flask-MakeStatic: detected change in foo, compiling\n'
                )

                with closing(app.send_static_file('foo')) as response:
                    self.assertEqual(response.status_code, 200)

    def test_static_view(self):
        with self.make_static('working') as (app, make_static):
            client = app.test_client()
            with closing(client.get('/static/foo')) as response:
                self.assertEqual(response.status_code, 200)

            with closing(client.get('/static/spam')) as response:
                self.assertEqual(response.status_code, 200)

            with closing(client.get('/static/bar')) as response:
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.data, b'abc\ndef\n')

            with closing(client.get('/static/baz')) as response:
                self.assertEqual(response.status_code, 404)

            with closing(client.get('/static/eggs.css')) as response:
                self.assertEqual(response.status_code, 200)

            with catch_warnings(record=True): # silence RuleMissing warning
                with catch_stdout() as stdout:
                    filename = os.path.basename(get_temporary_filename(
                        directory=make_static.assets_folder
                    ))
                    time.sleep(0.05)

            self.assertEqual(
                stdout.getvalue(),
                'Flask-MakeStatic: detected new asset %s, compiling\n' %
                filename
            )

            with closing(client.get('/static/' + filename)) as response:
                self.assertEqual(response.status_code, 404)

            with catch_stdout() as stdout:
                bump_modification_time(
                    os.path.join(make_static.assets_folder, 'foo')
                )
                time.sleep(0.05)

            self.assertEqual(
                stdout.getvalue(),
                'Flask-MakeStatic: detected change in foo, compiling\n'
            )

            with closing(client.get('/static/foo')) as response:
                self.assertEqual(response.status_code, 200)

    def test_compile(self):
        app = Flask('working')
        make_static = MakeStatic(app)
        with app.app_context():
            make_static.compile()

        # ensure that app is not patched and assets are not compiled on demand
        app = Flask('working')
        client = app.test_client()
        with closing(client.get('/static/foo')) as response:
            self.assertEqual(response.status_code, 200)

        with closing(client.get('/static/spam')) as response:
            self.assertEqual(response.status_code, 200)

        with closing(client.get('/static/bar')) as response:
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data, b'abc\ndef\n')

        with closing(client.get('/static/eggs.css')) as response:
            self.assertEqual(response.status_code, 200)

    def test_compile_warns_on_missing_rule(self):
        app = Flask('missing_rule')
        make_static = MakeStatic(app)
        with catch_warnings(record=True) as warnings:
            with app.app_context():
                make_static.compile()
        self.assertEqual(len(warnings), 1)
        self.assertTrue(issubclass(warnings[-1].category, RuleMissing))
        self.assertEqual(str(warnings[-1].message),
                         'cannot find a rule for foo.css')


class WatcherTestCase(unittest.TestCase):
    def assert_(self, **kwargs):
        self.assertEqual(self.added_files, kwargs.pop('added_files', []))
        del self.added_files[:]
        self.assertEqual(self.modified_files, kwargs.pop('modified_files', []))
        del self.modified_files[:]
        self.assertEqual(self.removed_files, kwargs.pop('removed_files', []))
        del self.removed_files[:]
        self.assertEqual(self.added_directories,
                         kwargs.pop('added_directories', []))
        del self.added_directories[:]
        self.assertEqual(self.modified_directories,
                         kwargs.pop('modified_directories', []))
        del self.modified_directories[:]
        self.assertEqual(self.removed_directories,
                         kwargs.pop('removed_directories', []))
        del self.removed_directories[:]

        if kwargs:
            raise TypeError(
                'unexpected keyword argument: %s' % kwargs.popitem()[0]
            )

    def test(self):
        watcher = ThreadedWatcher()
        self.added_files = []
        self.modified_files = []
        self.removed_files = []
        self.added_directories = []
        self.modified_directories = []
        self.removed_directories = []
        watcher.file_added.connect(self.added_files.append)
        watcher.file_modified.connect(self.modified_files.append)
        watcher.file_removed.connect(self.removed_files.append)
        watcher.directory_added.connect(self.added_directories.append)
        watcher.directory_modified.connect(self.modified_directories.append)
        watcher.directory_removed.connect(self.removed_directories.append)

        directory = get_temporary_directory()
        watcher.add_directory(directory)
        watcher.watch(sleep=0.01)

        foo = os.path.join(directory, 'foo')
        open(foo, 'w').close()
        time.sleep(0.05) # reasonable amount of reaction time
        self.assert_(added_files=[foo], modified_directories=[directory])

        bump_modification_time(foo)
        time.sleep(0.05)
        self.assert_(modified_files=[foo], modified_directories=[directory])

        os.remove(foo)
        time.sleep(0.05)
        self.assert_(removed_files=[foo], modified_directories=[directory])

        bar = os.path.join(directory, 'bar')
        os.mkdir(bar)
        time.sleep(0.05)
        self.assert_(added_directories=[bar], modified_directories=[directory])

        baz = os.path.join(bar, 'baz')
        open(baz, 'w').close()
        time.sleep(0.05)
        self.assert_(added_files=[baz], modified_directories=[bar])

        bump_modification_time(baz)
        time.sleep(0.05)
        self.assert_(modified_files=[baz], modified_directories=[bar])

        os.remove(baz)
        time.sleep(0.05)
        self.assert_(removed_files=[baz], modified_directories=[bar])

        os.rmdir(bar)
        time.sleep(0.05)
        self.assert_(removed_directories=[bar],
                     modified_directories=[directory])

        watcher.stop()
        time.sleep(0.05)
        self.assertEqual(threading.active_count(), 1)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MakeStaticTestCase))
    suite.addTest(unittest.makeSuite(WatcherTestCase))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
"""
    flask.ext.makestatic
    ~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2013 by Daniel Neuhäuser
    :license: BSD, see LICENSE.rst for details
"""
import os
import re
import warnings
import subprocess

from flask.ext.makestatic.watcher import ThreadedWatcher


__version__ = '0.1.0'
__version_info__ = (0, 1, 0)


_section_re = re.compile(r"\[(?P<file_re>[^\]]+)\]")
_command_re = re.compile(r'\s*(?P<command>.*?)\s*$')


class RuleMissing(Warning):
    """
    Warning that is emitted if a rule cannot be found.
    """


class ParsingError(Exception):
    def __init__(self, message, line, lineno):
        Exception.__init__(self, message, line, lineno)
        self.message = message
        self.line = line
        self.lineno = lineno


class _ConfigParser(object):
    def __init__(self, file):
        self.file = file
        self.lines = list(self.stripped_comments(enumerate(file, start=1)))

    def stripped_comments(self, lines):
        for lineno, line in lines:
            if '#' in line:
                comment_start = line.index('#')
                line = line[:comment_start].rstrip()
                if not line:
                    continue
            elif not line.strip():
                continue
            yield lineno, line

    def next_line(self):
        return self.lines.pop(0)

    def parse_rule(self):
        try:
            lineno, line = self.next_line()
        except IndexError:
            return
        match = _section_re.match(line)
        if match is None:
            raise ParsingError('expected new rule', line, lineno)
        regex = match.group('file_re')
        commands = []
        insert = False
        while True:
            try:
                lineno, line = self.next_line()
            except IndexError:
                break
            if _section_re.match(line):
                insert = True
                break
            commands.append(line.strip())
        if insert:
            self.lines.insert(0, (lineno, line))
        return regex, commands

    def parse(self):
        while True:
            result = self.parse_rule()
            if result is None:
                break
            yield result


class MakeStatic(object):
    """
    This class provides the extension interface. You can use it by calling it
    directly with a :class:`flask.Flask` instance::

        app = Flask(__name__)
        make_static = MakeStatic(app)

    Once initialized :class:`MakeStatic` will parse the `assets.cfg`
    configuration file corresponding to the initialized application.
    """
    def __init__(self, app):
        self.app = app

        with self.app.open_resource('assets.cfg', 'r') as config_file:
            self.matcher, self.rulesets = self.parse_config(config_file)

    @property
    def assets_folder(self):
        return os.path.join(self.app.root_path, 'assets')

    def parse_config(self, config_file):
        file_regexes = []
        rulesets = []
        for file_regex, rules in _ConfigParser(config_file).parse():
            file_regexes.append('(%s)' % file_regex)
            rulesets.append(rules)
        matcher = re.compile('^%s$' % '|'.join(file_regexes)).match
        return matcher, rulesets

    def get_rules(self, filename):
        match = self.matcher(filename)
        if match:
            return self.rulesets[match.lastindex - 1]

    def watch(self, sleep=0.1):
        """
        Starts a daemon thread that watches the `static` directory for changes
        and compiles the files that do change.

        Returns a :class:`ThreadedWatcher`, if you want to turn of watching you
        can call :meth:`ThreadedWatcher.stop`.

        When run in a process started by the reloader, this does nothing to
        prevent the start of an unnecessary second watcher.

        :param sleep: The amount of time in seconds that should be slept
                      between checks for changes, may be ignored.
        """
        if os.environ.get('WERKZEUG_RUN_MAIN') == 'true':
            # We are running in the Werkzeug reloader, which runs the setup
            # code twice (don't ask) which means we are called twice. This
            # means that the Watcher is also created twice and we compile all
            # assets twice. That is annoying to say the least, so we just
            # return and do nothing.
            return
        watcher = ThreadedWatcher()
        @watcher.file_added.connect
        def on_file_added(filename):
            print(
                u'Flask-MakeStatic: detected new asset %s, compiling' %
                os.path.relpath(filename, self.assets_folder)
            )
            self.compile_asset(filename)
        @watcher.file_modified.connect
        def on_file_modified(filename):
            print(
                u'Flask-MakeStatic: detected change in %s, compiling' %
                os.path.relpath(filename, self.assets_folder)
            )
            self.compile_asset(filename)
        watcher.add_directory(self.assets_folder)
        watcher.watch(sleep=sleep)
        self.compile() # initial compile
        return watcher

    def compile(self):
        """
        Compiles all assets to static files in one go.

        This works only when done within an application context of an
        initialized application.

        Emits a :class:`RuleMissing` warning for each file in `assets` for
        which no rule exists.
        """
        for root, _, files in os.walk(self.assets_folder):
            for file in files:
                self.compile_asset(os.path.join(root, file))

    def compile_asset(self, filename):
        relative_filename = os.path.relpath(filename, self.assets_folder)
        rules = self.get_rules(relative_filename)
        if rules is None:
            warnings.warn(
                'cannot find a rule for %s' % relative_filename,
                RuleMissing,
            )
        else:
            static = os.path.join(self.app.static_folder, relative_filename)
            static_base = os.path.splitext(static)[0]
            for rule in rules:
                subprocess.check_call(
                    rule.format(
                        asset=filename,
                        static=static,
                        static_dir=self.app.static_folder,
                        static_base=static_base
                    ),
                    shell=True
                )


__all__ = ['MakeStatic']



# coding: utf-8
import os

from setuptools import setup


def get_version():
    # We can't import flask.ext.makestatic because that would require all
    # dependencies to be installed.
    init_path = os.path.join(
        os.path.dirname(__file__), 'flask_makestatic', '__init__.py'
    )
    with open(init_path) as init_file:
        for line in init_file:
            if line.startswith('__version__'):
                return line.split('=')[1].replace("'", '').strip()
        raise ValueError('__version__ not found in %s' % init_path)


def get_long_description():
    with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
        return readme.read()


setup(
    name='Flask-MakeStatic',
    version=get_version(),
    url='https://github.com/DasIch/Flask-MakeStatic',
    license='BSD',
    author='Daniel Neuhäuser',
    author_email='ich@danielneuhaeuser.de',
    description='Make for your flask app assets',
    long_description=get_long_description(),
    packages=['flask_makestatic'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=['Flask>=0.9'],
    test_suite='test_makestatic.suite',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
